#!/usr/bin/env python
#Once upon a time...

import pdb,sys,os,argparse
import pysam
import time
import parabam
import gzip

import time

import Queue as Queue2

from collections import Counter
from multiprocessing import Queue,Process
from itertools import izip

class MergePackage(parabam.core.Package):
    def __init__(self,object results,str subset_type):
        super(MergePackage,self).__init__(results)
        self.subset_type = subset_type

class Handler(parabam.core.Handler):

    def __init__(self,object parent_bam, object output_paths,object inqu,
                      object constants,object pause_qus,dict out_qu_dict):

        super(Handler,self).__init__(parent_bam,output_paths,inqu,constants,pause_qus,
                                        out_qu_dict,report=False)
        
        self._user_subsets = list(constants.user_subsets)
        self._out_file_objects = self.__get_out_file_objects__()
        self._merged = 0

    def __get_out_file_objects__(self):
        file_objects = {}
        output_paths = self._output_paths

        file_objects = {}
        for subset in self._user_subsets:
            output_path = output_paths[self._parent_bam.filename][subset]
            merge_type = self.__get_subset_merge_type__(output_path)
            cur_file_obj = self.__get_file_for_merge_type__(merge_type,output_path)
            file_objects[subset] = cur_file_obj

        return file_objects

    def __get_file_for_merge_type__(self,merge,path):
        if merge == "gzip":
            return gzip.open(path,"wb")
        elif merge == "txt":
            return open(path,"w")
        return pysam.AlignmentFile(path,"wb",header=self._parent_bam.header)

    def __get_subset_merge_type__(self,path):
        root,extension = os.path.splitext(path)
        if extension == ".gzip" or extension == ".gz":
            return "gzip"
        elif extension == ".txt":
            return "txt"
        return "pysam"

    def __periodic_action__(self,iterations):
        if self._destroy:
            try:
                pack = self._inqu.get(False,10)
                self._inqu.put(pack)
            except Queue2.Empty:
                self._finished = True

    def __new_package_action__(self,new_package,**kwargs):
        #Handle the result. Result will always be of type parabam.support.MergePackage
        subset_type = new_package.subset_type
        for merge_count,merge_path in new_package.results:
            if merge_count > 0:
                for item in self.__get_entries_from_file__(merge_path,subset_type):
                    self._out_file_objects[subset_type].write(item)
            os.remove(merge_path)
            self._merged += 1
        time.sleep(.5)

    def __get_entries_from_file__(self,path,subset):
        merge_type = self.__get_subset_merge_type__(path)
        if merge_type == "pysam":
            try:
                file_object = pysam.AlignmentFile(path,"rb")
            except IOError:
                print "fail on path %s" % (path,)
                return
            for read in file_object.fetch(until_eof=True):
                yield read
        elif merge_type == "gzip":
            file_object = gzip.open(path,"rb")
            for item in file_object:
                yield item
        elif merge_type == "txt":
            file_object = open(path,"r")
            for item in file_object:
                yield item
        else:
            print "[Error] Unrecognised merge type"
            return
        file_object.close()

    def __get_unique_temp_path__(self,temp_type,temp_dir="."):
        return "%s/%sTEMP%d.bam" % (temp_dir,temp_type,int(time.time()),)

    def __close_all_out_files__(self):
        for subset,file_obj in self._out_file_objects.items():
            file_obj.close()
        del self._out_file_objects

    def __handler_exit__(self,**kwargs):
        self.__close_all_out_files__()
        
#...happily ever after



import time
import sys
import os
import gc
import shutil
import gzip

import pysam
import parabam

import Queue as Queue2

from itertools import izip
from multiprocessing import Queue,Process
from abc import ABCMeta, abstractmethod

#TODO: error handle incorrect return from user engine

class Task(parabam.core.Task):
    __metaclass__=ABCMeta

    def __init__(self,parent_bam,inqu,outqu,statusqu,task_size,constants):
        super(Task, self).__init__(parent_bam=parent_bam,
                                    inqu=inqu,
                                    outqu=outqu,
                                    statusqu=statusqu,
                                    task_size=task_size,
                                    constants=constants)

        self._engine = constants.user_engine
        self._user_constants = constants.user_constants

    @abstractmethod
    def __handle_engine_output__(self,engine_output,read):
        pass

    def __pre_run_routine__(self,iterator,**kwargs):
        self._system = {}

    def __post_run_routine__(self,**kwargs):
        pass

    def __process_task_set__(self,iterator):
        engine = self._engine
        next_read = iterator.next 
        parent_bam = self._parent_bam
        handle_output = self.__handle_engine_output__
        user_constants = self._user_constants
        
        #StopIteration caught in parabam.core.Task.run
        for i in xrange(self._task_size):    
            read = next_read()
            engine_output = engine(read,user_constants,parent_bam)
            handle_output(engine_output,read)

    def __get_extension__(self,path):
        root,extension = os.path.splitext(path)
        return extension

class PairTask(Task):
    __metaclass__ = ABCMeta

    def __init__(self,parent_bam,inqu,outqu,statusqu,task_size,constants):
        super(PairTask, self).__init__(parent_bam=parent_bam,
                                    inqu=inqu,
                                    outqu=outqu,
                                    statusqu=statusqu,
                                    task_size=task_size,
                                    constants=constants)

        if constants.include_duplicates:
            self.__read_filter__ = self.__filter__
        else:
            self.__read_filter__ = self.__filter_duplicates__

    @abstractmethod
    def __handle_engine_output__(self,engine_output,read):
        pass

    def __pre_run_routine__(self,iterator):
        super(PairTask,self).__pre_run_routine__(iterator)
        self._loners = {}

    def __post_run_routine__(self,**kwargs):
        super(PairTask,self).__post_run_routine__()
        self.__stash_loners__(self._loners)
        del self._loners

    def __process_task_set__(self,iterator):
        engine = self._engine
        next_read = iterator.next 
        parent_bam = self._parent_bam
        query_loners = self.__query_loners__ 
        cdef int size = self._task_size
        handle_output = self.__handle_engine_output__
        user_constants = self._user_constants
        counts = self._counts

        loners = self._loners

        #StopIteration caught in parabam.core.Task.run
        for i in xrange(size):
            read = next_read()
            read1,read2 = query_loners(read,loners)

            if read1:
                engine_output = engine((read1,read2),user_constants,parent_bam)
                handle_output(engine_output,(read1,read2,))

    def __stash_loners__(self,loners):
        loner_count = 0
        loner_path = self.__get_temp_path__("chaser")
        loner_file = pysam.AlignmentFile(loner_path,"wb",header=self._parent_bam.header)

        for qname,read in loners.items():
            loner_count += 1
            loner_file.write(read)
        self._system["chaser"] = (loner_count,loner_path,)
        loner_file.close()

    def __query_loners__(self,read,loners):
        if self.__read_filter__(read):#Pair processing only handles primary pairs
            return None, None
        try:
            mate = loners[read.qname]
            del loners[read.qname]
            return read,mate
        except KeyError:
            loners[read.qname] = read
            return None,None

    def __filter__(self,read):
        return read.is_secondary or (read.flag & 2048 == 2048)

    def __filter_duplicates__(self,read):
        return read.is_secondary or (read.flag & 2048 == 2048) or read.is_duplicate

class Handler(parabam.core.Handler):
    __metaclass__=ABCMeta

    def __init__(self,object parent_bam, object output_paths,object inqu,
                object constants,object pause_qus,dict out_qu_dict,object report=True):
        
        super(Handler,self).__init__(parent_bam = parent_bam,output_paths = output_paths,
                                     inqu=inqu,constants=constants,pause_qus=pause_qus,
                                     out_qu_dict=out_qu_dict)

        self._system_subsets = constants.system_subsets

        #Stage stores serve as a staging area
        #for packages to be sent on to subhandlers
        self._stage_stores = {} 
        for subset in self._system_subsets:
            self._stage_stores[subset] = []

        self._out_qu_dict = out_qu_dict
        self._stagecount = 0

    #Child classes MUST call super
    def __new_package_action__(self,new_package,**kwargs):
        results = new_package.results
        self.__auto_handle__(results,self._stats)

        if "system" in results.keys():
            self.__add_system_to_stage__(results["system"])
        
    def __add_system_to_stage__(self,system_results):
        for subset,(count,path) in system_results.items():
             self._stage_stores[subset].append((count,path))

    #Child classes MUST call super
    def __periodic_action__(self,iterations):
        self.__decide_if_finished__()
        for subset in self._system_subsets:
            if self.__test_stage_store__(subset):
                self.__add_staged_system_task__(results=self._stage_stores[subset],
                                        subset_type=subset)
                self.__update_stage_store__(subset)

    def __decide_if_finished__(self):
        if self._destroy:
            has_stage_tasks = False
            for subset,staged in self._stage_stores.items():
                if len(staged) > 0:
                    has_stage_tasks = True 
                    break
            if not has_stage_tasks:
                self._finished = True
                    
    def __update_stage_store__(self,subset):
        self._stagecount += 1
        self._stage_stores[subset] = []
        gc.collect()

    def __test_stage_store__(self,subset):
        if (not self._processing) or self._destroy:
            return True
        else:
            return len(self._stage_stores[subset]) > 20
    
    def __add_staged_system_task__(self,results,subset_type):
        if subset_type == "chaser":
            res = parabam.chaser.ChaserResults(results=results,
                                chaser_type="origin")
            self._out_qu_dict["chaser"].put(res)

    #Child classes MUST call super
    def __handler_exit__(self, **kwargs):
        if self._verbose:
            self.__standard_output__("[Status] Processing complete. %d reads processed" \
                                        % (self.__total_reads__(),))

class Interface(parabam.core.Interface):

    __metaclass__ = ABCMeta

    def __init__(self,temp_dir):
        super(Interface,self).__init__(temp_dir)

    def __get_module_and_vitals__(self,code_path):
        if os.getcwd not in sys.path:
            sys.path.append(os.getcwd())

        code_path = code_path.replace(".py","")

        try:
            module = __import__(code_path, fromlist=[''])
        except ImportError:
            sys.stderr.write("[Error] parabam can't find user specified instructions\n"\
                             "\tEnsure instruction code is in current working directory\n")
            raise SystemExit

        user_engine = module.engine
        user_constants = {}
        if hasattr(module,"set_constants"):
            module.set_constants(user_constants)

        return module,user_engine,user_constants

    @abstractmethod
    def __get_handler_bundle__(self,**kwargs):
        pass

    @abstractmethod
    def __get_destroy_handler_order__(self):
        '''Should return a list of hanlder classes in the 
        order that they are to be destroyed by the leviathon
        all hanlder class should have a corresponding bundle entry'''
        pass 

    @abstractmethod
    def __get_output_paths__(self,input_paths,**kwargs):
        pass

    @abstractmethod
    def __get_task_class__(self,**kwargs):
        pass

    @abstractmethod
    def __get_queue_names__(self,**kwargs):
        '''Predefine the names of queues in this implementation
        these names are converted to queue objects by the leviathon
        use these names when reference queus in processor and handler
        bundles'''
        pass

    @abstractmethod
    def __get_extra_const_args__(self,**kwargs):
        '''return empty dict if not needed'''
        pass

    def __get_const_args__(self,**kwargs):
        '''Any function overwriting this one MUST make a super call to this function.
        i.e args = super(<class_name>,self).__get_const_args__(**kwargs)'''

        args = {}
        args["temp_dir"] = self._temp_dir


        defaults = ["total_procs","task_size","verbose","user_constants",
                    "reader_n","user_constants", "user_engine", "fetch_region", 
                    "pair_process", "include_duplicates","debug"]

        for key,val in kwargs.items():
            if key in defaults:
                args[key] = val

        if args["pair_process"]:
            args["system_subsets"] = ["chaser"]
        else:
            args["system_subsets"] = []

        return args

    def __update_final_output_paths__(self,input_path,output_paths,final_output_paths):
        if "global" in output_paths.keys():
            for path in output_paths["global"]:
                if path not in final_output_paths["global"]:
                    final_output_paths["global"].append(path)

        final_output_paths[input_path] = []
        if input_path in output_paths.keys():
            output_path = output_paths[input_path]
            if type(output_path) == dict:
                for subset,path in output_path.items():
                    final_output_paths[input_path].append(path)    
            elif type(output_path) == list:
                final_output_paths[input_path] = output_path
            elif type(output_path) == str:
                final_output_paths[input_path].append(output_path)
            else:
                sys.stderr.write("[Error] Unexpected output path type: %s \n" \
                    % (type(output_path)))
                raise SystemExit

    def __output_files_to_cwd__(self,final_output_paths):
        revised_output_paths = {}
        for master_path,out_paths in final_output_paths.items():
            revised_output_paths[ master_path ] = []
            for path in out_paths:
                head,tail = os.path.split(path)
                new_path = os.path.abspath(os.path.join(".",tail))
                real_path = self.__move_output_file__(path,new_path)
                revised_output_paths[master_path].append(real_path)
        return revised_output_paths

    def __move_output_file__(self,current_path,new_path):
        try:
            shutil.move(current_path,new_path)
            return new_path
        except shutil.Error,e: 
            root,ext = os.path.splitext( new_path )
            alternate_path = root + "_%d" % time.time() + ext

            sys.stderr.write("[Warning] Something went wrong when copying output to working directory.\n")
            sys.stdout.write("[Update]Trying to create output using unique filename:\n")
            sys.stdout.write("\t\t%s\n" % alternate_path)

            shutil.move(current_path,alternate_path)
            return alternate_path

    def __report_file_names__(self,final_output_paths,input_paths):
        sys.stdout.write("[Status] This run will output the following files:\n")
        for master_path,child_paths in final_output_paths.items():
            if master_path in input_paths:
                for path in child_paths:
                    root,name = os.path.split(path)
                    sys.stdout.write("\t %s\n" % (name,))            
        sys.stdout.write("\n")

    def run(self,**kwargs):

        input_paths = kwargs["input_bams"]

        const_args = self.__get_const_args__(**kwargs)
        const_args.update(self.__get_extra_const_args__(**kwargs))
        constants = parabam.core.Constants(**const_args)

        task_class = self.__get_task_class__(**kwargs)
        queue_names = self.__get_queue_names__(**kwargs)

        handler_order = self.__get_destroy_handler_order__()
        handler_bundle = self.__get_handler_bundle__(**kwargs)

        if constants.pair_process:
            self.__prepare_for_pair_processing__(handler_bundle,handler_order,
                                                 queue_names,constants,task_class)
        
        update_interval = self.__get_update_interval__(constants.verbose)

        leviathon = parabam.core.Leviathon(constants,handler_bundle,
                                           handler_order,queue_names,
                                           update_interval,task_class)

        final_output_paths = {"global":[]}
        global_paths = self.__get_global_output_path__(**kwargs)

        for input_path in input_paths:
            output_paths = self.__get_output_paths__(input_path=input_path,**kwargs)
            self.__global_paths_to_output__(output_paths,global_paths,**kwargs)
            self.__update_final_output_paths__(input_path,output_paths,
                                               final_output_paths)

            if constants.verbose and not kwargs["keep_in_temp"]: 
                self.__report_file_names__(final_output_paths,input_path)

            leviathon.run(input_path,output_paths)

        if not kwargs["keep_in_temp"]:
            final_output_paths = self.__output_files_to_cwd__(final_output_paths)
        
        self.__remove_empty_entries__(final_output_paths)
        return final_output_paths

    def __get_global_output_path__(self,**kwargs):
        return {}

    def __global_paths_to_output__(self,output_paths,global_paths,**kwargs):
        if not global_paths == {}:
            output_paths["global"] = global_paths["global"]

    def __get_global_output__(self,**kwargs):
        return {}

    def __get_update_interval__(self,verbose):
        if verbose == 1: 
            return 7500
        else:
            return 100

    def __remove_empty_entries__(self,final_output_paths):
        for master_path,child_paths in final_output_paths.items():
            if len(child_paths) == 0:
                del final_output_paths[master_path]

    def __prepare_for_pair_processing__(self,handler_bundle,handler_order,
                                            queue_names,constants,Task):        
        handler_bundle[parabam.chaser.Handler] = {"inqu":"chaser",
                                                  "constants":constants,
                                                  "out_qu_dict":["main"],
                                                  "Task":Task}

        for handler_class,handler_args in handler_bundle.items():
            if issubclass(handler_class,Handler):
                handler_bundle[handler_class]["out_qu_dict"].append("chaser")
        queue_names.append("chaser")
        handler_order.insert(0,parabam.chaser.Handler)
        constants.total_procs = constants.total_procs / 2

    def default_parser(self):
        parser = super(Interface,self).default_parser()

        parser.add_argument('--instruc','-i',metavar='INSTRUCTION',required=True
            ,help='The instruction file, written in python,\n'\
            'that we wish to apply to the input BAM.')
        parser.add_argument('--input','-b',metavar='INPUT', nargs='+',required=True
            ,help='The file(s) we wish to operate on.\n'\
            ' Multiple entries should be separated by a single space')
        parser.add_argument('--debug',action="store_true",default=False,
            help="Only the first 5million reads will be processed")
        parser.add_argument('--pair',action="store_true",default=False
            ,help="A pair processor is used instead of a conventional processor")
        parser.add_argument('-r','--region',type=str,metavar="REGION",nargs='?',default=None
            ,help="The process will be run only on reads from\n"\
            "this region. Regions should be colon separated as\n"\
            "specified by samtools (eg \'chr1:1000,5000\')")
        parser.add_argument('-d',action="store_false",default=True,
            help="parabam will not process reads marked duplicate.")
        parser.add_argument('-v', choices=[0,1,2],default=0,type=int,
            help="Indicate the amount of information output by the program:\n"\
            "\t0: No output [Default]\n"\
            "\t1: Total Reads Processed\n"\
            "\t2: Detailed output")

        return parser
        
    @abstractmethod
    def get_parser(self):
        pass

#...happily ever after



#Once upon a time...
import pysam
import parabam
import shutil
import time
import sys
import os
import copy
import numpy as np

from itertools import izip
from multiprocessing import Queue
from abc import ABCMeta, abstractmethod

class StatCore(object):

    def __init__(self,object constants):
        self._constants
        self._counts = {}
        self._local_structures = {}
        self._system = {}

    def __pre_run_routine__(self,iterator,**kwargs):
        super(StatCore,self).__pre_run_routine__(iterator)
        for name,structure in self._constants.user_structures.items():
            self._local_structures[name] = structure.empty_clone()
            self._counts[name] = 0

    def __get_results__(self):
        results = {}
        results["structures"] = self.__unpack_structures__(self._local_structures)
        results["counts"] = self._counts
        results["system"] = self._system
        return results

    def __post_run_routine__(self,**kwargs):
        super(StatCore,self).__post_run_routine__()
        pass

    def __handle_engine_output__(self,engine_output,read):
        if engine_output:#Allows return False
            local_structures = self._local_structures
            for name,package in engine_output.items():
                self._counts[name] += 1
                local_structures[name].add(**package)

    def __unpack_structures__(self,structures):
        unpacked = []
        for name,struc in structures.items():
            unpacked.append( (name,struc.data) )
        return unpacked

class Task(StatCore,parabam.command.Task):

    def __init__(self,parent_bam,inqu,outqu,statusqu,task_size,constants):
        
        parabam.command.Task.__init__(self,parent_bam=parent_bam,
                                            inqu=inqu,
                                            outqu=outqu,
                                            statusqu=statusqu,
                                            task_size=task_size,
                                            constants=constants)
        StatCore.__init__(self,constants)
        
class PairTask(StatCore,parabam.command.PairTask):

    def __init__(self,parent_bam,inqu,outqu,statusqu,task_size,constants):
        
        parabam.command.PairTask.__init__(self,parent_bam=parent_bam,
                                            inqu=inqu,
                                            outqu=outqu,
                                            statusqu=statusqu,
                                            task_size=task_size,
                                            constants=constants)
        StatCore.__init__(self,constants)


class Handler(parabam.command.Handler):

    def __init__(self,object parent_bam, object output_paths,object inqu,
                object constants,object pause_qus,dict out_qu_dict,object report=True):

        super(Handler,self).__init__(parent_bam = parent_bam,output_paths = output_paths,
                                     inqu=inqu,constants=constants,pause_qus=pause_qus,
                                     out_qu_dict=out_qu_dict)

        self._final_structures = {}
        for name,struc in self._constants.user_structures.items():
            self._final_structures[struc.name] = struc.empty_clone()

    def __new_package_action__(self,new_package,**kwargs):
        super(Handler,self).__new_package_action__(new_package)
        results = new_package.results
        for name,data in results["structures"]:
            final_struc = self._final_structures[name]
            final_struc.merge(data)

    def __handler_exit__(self):
        super(Handler,self).__handler_exit__()
        constants = self._constants

        #Append to global csv
        if constants.analysis_names: #Check that there are global analyses
            data_str = self.__get_data_str_from_names__(constants.analysis_names,self._final_structures)
            with open(self._output_paths["global"][0],"a") as out_object:
                out_object.write("%s%s\n" % (self._parent_bam.filename,data_str))
        
        #Output numpy arrays
        for name,structure in self._final_structures.items():
            if structure.struc_type == np.ndarray:
                structure.write_to_csv(self._output_paths[self._parent_bam.filename][name])

    def __get_data_str_from_names__(self,names,user_structures):
        data_str = ""
        for name in names:
            if user_structures[name].struc_type == np.ndarray:
                continue
            cur_data = user_structures[name].data
            data_str += ",%.5f" % (cur_data,)
        return data_str

class UserStructure(object):

    __metaclass__ = ABCMeta

    def __init__(self,name,struc_type,store_method,data):
        self.struc_type = struc_type
        self.store_method = store_method 
        self.data = data
        self.org_data = copy.copy(data)
        self.name = name

        if store_method == "max":
            self.add = self.add_max
            self.merge = self.merge_max
        elif store_method == "min":
            self.add = self.add_min
            self.merge = self.merge_min
            self.data = float('inf')
            self.org_data = copy.copy(float('inf'))
        else:
            self.add = self.add_cumu
            self.merge = self.merge_cumu

    def max_decision(self,result,existing):
        return max(result,existing)

    def min_decision(self,result,exisiting):
        return min(result,exisiting)

    @abstractmethod
    def empty_clone(self):
        pass

    @abstractmethod
    def add_max(self,result):
        pass

    @abstractmethod
    def add_min(self,result):
        pass

    @abstractmethod
    def add_cumu(self,result):
        pass

    @abstractmethod
    def merge_max(self,result):
        pass

    @abstractmethod
    def merge_min(self,result):
        pass

    @abstractmethod
    def merge_cumu(self,result):
        pass

class NumericStructure(UserStructure):
    def __init__(self,name,struc_type,store_method,data,log_scaling=False):
        super(NumericStructure,self).__init__(name,struc_type,store_method,data)
        self.log_scaling = log_scaling

    def empty_clone(self):
        return NumericStructure(self.name,self.struc_type,self.store_method,self.org_data)

    def add_cumu(self,result):
        self.data += result

    def add_max(self,result):
        self.data = self.max_decision(result,self.data)

    def add_min(self,result):
        self.data = self.min_decision(result,self.data)

    def merge_max(self,result):
        self.data = self.max_decision(self.data,result)
        del result

    def merge_min(self,result):
        self.data = self.min_decision(self.data,result)
        del result

    def merge_cumu(self,result):
        self.data += result
        del result

class ArrayStructure(UserStructure):
    def __init__(self,name,struc_type,store_method,data):
        super(ArrayStructure,self).__init__(name,struc_type,store_method,data)

    def empty_clone(self):
        return ArrayStructure(self.name,self.struc_type,self.store_method,self.org_data)

    def add_max(self,result,coords):
        existing = self.data[coords]
        self.data[coords] = self.max_decision(result,existing)

    def add_min(self,result,coords):
        existing = self.data[coords]
        self.data[coords] = self.min_decision(result,existing)

    def add_cumu(self,result):
        self.data += result

    def merge_max(self,result):
        self.data = np.maximum(self.data,result)
        del result

    def merge_min(self,result):
        self.data = np.minimum(self.data,result)
        del result

    def merge_cumu(self,result):
        self.data += result
        del result

    def write_to_csv(self,out_path):
        np.savetxt(out_path,self.data,fmt="%.3f",delimiter=",")

class Interface(parabam.command.Interface):

    def __init__(self,temp_dir):
        super(Interface,self).__init__(temp_dir)
    
    def run_cmd(self,parser):

        cmd_args = parser.parse_args()

        module,user_engine,user_constants = self.__get_module_and_vitals__(cmd_args.instruc)
        user_struc_blueprint = {}
        module.set_structures(user_struc_blueprint)

        self.run(input_bams=cmd_args.input,
            total_procs = cmd_args.p,
            task_size = cmd_args.s,
            verbose= cmd_args.v,
            reader_n = cmd_args.f,
            user_constants = user_constants,
            user_struc_blueprint = user_struc_blueprint,
            user_engine = user_engine,
            fetch_region = cmd_args.region,
            pair_process=cmd_args.pair,
            include_duplicates=cmd_args.d,
            debug = cmd_args.debug,
            announce = True)


    def run(self,input_bams,total_procs,task_size,user_constants,user_engine,
            user_struc_blueprint,user_specified_outpath=None,
            reader_n = 2,fetch_region=None,side_by_side=2,
            keep_in_temp=False,verbose=0,pair_process=False,
            include_duplicates=True,debug=False,announce=False):

        ''' Docstring! '''
        args = dict(locals())
        del args["self"]

        if not verbose:
            announce = False
        self.__introduce__("parabam stat",announce)

        results = super(Interface,self).run(**args)

        self.__goodbye__("parabam stat",announce)
        return results

    def __get_destroy_handler_order__(self):
        return [Handler]

    def __get_queue_names__(self,**kwargs):
        return ["main"]

    def __get_extra_const_args__(self,user_struc_blueprint,**kwargs):
        user_structures = self.__create_structures__(user_struc_blueprint)
        analysis_names = self.__get_non_array_names__(user_struc_blueprint)

        return {"user_structures":user_structures,
                "analysis_names":analysis_names}

    def __get_handler_bundle__(self,**kwargs):
        handler_bundle = { Handler: {"inqu":"main",
                                    "out_qu_dict":[]}}
        return handler_bundle


    def __get_global_output_path__(self,user_struc_blueprint,user_specified_outpath,**kwargs):
        analysis_names = self.__get_non_array_names__(user_struc_blueprint)
        if analysis_names:
            if not user_specified_outpath:
                global_filename = os.path.join(".",self._temp_dir,"parabam_stat_%d_%d.csv"\
                                         % (time.time(),os.getpid()))
            else:
                global_filename = user_specified_outpath
            self.__create_output_files__(global_filename,analysis_names)
            return {"global": [global_filename]}
        else:
            return {"global": []}

    def __get_output_paths__(self,input_path,user_specified_outpath,
                             user_struc_blueprint,
                             **kwargs):
        output_paths = {input_path:{}}
        for name,blueprint in user_struc_blueprint.items():
            if type(blueprint["data"]) == np.ndarray:
                path_id,ext = os.path.splitext(os.path.basename(input_path))
                csv_path = "%s_%s.csv" % (path_id,name,)
                output_paths[input_path][name] = os.path.join(".",self._temp_dir,csv_path)
        return output_paths

    def __get_task_class__(self,pair_process,**kwargs):
        if pair_process:
            return PairTask
        else:
            return Task

    def __get_queues__(self,object constants,**kwargs):
        queues = {"main":Queue()}
        return queues

    def __create_output_files__(self,output_path,analysis_names):
        header = "Sample,%s\n" % (",".join(analysis_names),)
        with open(output_path,"w") as out_obj:
            out_obj.write(header)

    def __get_non_array_names__(self,user_struc_blueprint):
        names = []
        for name,blueprint in user_struc_blueprint.items():
            if not type(blueprint["data"]) == np.ndarray:
                names.append(name)
        names.sort()
        return names

    def __create_structures__(self,user_struc_blueprint):
        user_structures = {}
        class_to_type_map = {int:NumericStructure,
                             float:NumericStructure,
                             np.ndarray:ArrayStructure}

        for name,definition in user_struc_blueprint.items():
            definition["name"] = name
            definition["struc_type"] = type(definition["data"])
            user_structures[name] = class_to_type_map[definition["struc_type"]](**definition)
        return user_structures

    def get_parser(self):
        #argparse imported in ./interface/parabam 
        parser = self.default_parser()

        parser.add_argument('--output','-o',metavar='OUTPUT', nargs='?',required=False
        ,help="Specify a name for the output CSV file. Only used with default `outmode`.\n"\
            "If this argument is not supplied, the output will take the following form:\n"\
            "parabam_stat_[UNIX_TIME].csv")

        return parser

#...happily ever after



import os
import time
import datetime
import sys
import Queue as Queue2
import gc
import shutil
import argparse

import pysam

from itertools import izip
from multiprocessing import Process,Queue
from abc import ABCMeta, abstractmethod

cdef class Handler:

    def __init__(self,object parent_bam, object output_paths,object inqu,
                object constants,object pause_qus,dict out_qu_dict,object report=True):

        self._parent_bam = parent_bam
        self._inqu = inqu
        self._pause_qus = pause_qus
        self._out_qu_dict = out_qu_dict

        self._report = report
        self._stats = {}

        self._constants = constants
        self._verbose = constants.verbose
        self._output_paths = output_paths

        self._periodic_interval = 10

        self._processing = True
        self._destroy = False
        self._finished = False

        if constants.verbose == 1:
            self._verbose = True
            self._update_output = self.__level_1_output__
        elif constants.verbose == 2 or constants.verbose == True:
            #In the case verbose is simply "True" or "level 2"
            self._verbose = True
            self._update_output = self.__level_2_output__
        else:#catching False and -v0
            self._verbose = False
            self._update_output = self.__level_2_output__

    def __standard_output__(self,out_str):
        sys.stdout.write(out_str + "\n")
        sys.stdout.flush()

    def __level_1_output__(self,out_str):
        #BUG! the fact this makes a call to __total_reads__ is ridiculous
        #this is making calls to a sub class method and just expecting it to be there.

        total_procd = self.__total_reads__()
        time = out_str.partition("Time: ")[2]
        sys.stdout.write("[Update] Processed: %d Time: %s\n" % (total_procd,time))
        sys.stdout.flush()

    def __level_2_output__(self,outstr):
        sys.stdout.write("\r" + outstr)
        sys.stdout.flush()

    #Must be overwritten if stats architecture is modififed
    def __total_reads__(self):
        if self._stats == {}:
            return 0
        return self._stats["total"]

    def listen(self,update_interval):
        destroy = self._destroy

        cdef int iterations = 0
        cdef int start_time = time.time()
        cdef int dealt  = 0

        update_output = self._update_output #speedup alias
        periodic_interval = self._periodic_interval #speedup alias
        finished = self.__is_finished__

        while not finished():
            iterations += 1
            #Listen for a process coming in...
            try:
                new_package = self._inqu.get(False)
                if type(new_package) == DestroyPackage:
                    self._destroy = True
                elif type(new_package) == EndProcPackage:
                    self._processing = False

                if not new_package.results == {}:#If results are present...
                    self.__new_package_action__(new_package) #Handle the results
                    dealt += 1

            except Queue2.Empty:
                #Queue empty. Continue with loop
                time.sleep(.5)

            if iterations % periodic_interval == 0: 
                self.__periodic_action__(iterations)

            if self._verbose and self._report and iterations % update_interval == 0:
                outstr = self.__format_update__(start_time)
                update_output(outstr)

        self._inqu.close()
        self.__handler_exit__()

    def __is_finished__(self):
        if not self._destroy or not self._finished:
            return False
        return True

    def __format_update__(self,start_time):
        stats = "[Update] "
        for name,value in self._stats.items():
            stats += "%s:%d " % (name,value)
        return "\r%s| Time: %d " %\
                (stats,self.__secs_since__(start_time),)

    #This code is a little ugly. Essentially, given a results
    #dictionary, it will go through and create a sensible output
    #string.
    def __auto_handle__(self,results,stats):
        for key_one in results:
            if type(results[key_one]) is int:
                if key_one in stats:
                    stats[key_one] += results[key_one]
                else:
                    stats[key_one] = results[key_one]
            elif type(results[key_one]) is dict:
                for key_two in results[key_one]:
                    if type(results[key_one][key_two]) is int:
                        if key_two in stats:
                            stats[key_two] += results[key_one][key_two]
                        else:
                            stats[key_two] = results[key_one][key_two]

    def __secs_since__(self,since):
        return int(time.time() - since)

    def __periodic_action__(self,iterations):
        #Overwrite with something that needs
        #to be done occasionally.       
        pass

    def __new_package_action__(self,new_package,**kwargs):
        #Handle the output of a task. Input will always be of type
        #parabam.Package. New results action should always call
        #self.__auto_handle__(new_package.results)
        pass

    def __handler_exit__(self,**kwargs):
        #When the handler finishes, do this
        pass

class Task(Process):
    __metaclass__=ABCMeta

    def __init__(self,parent_bam,inqu,outqu,task_size,constants):
        super(Task,self).__init__()
        self._parent_bam = parent_bam
        self._parent_path = self._parent_bam.filename
        self._header = self._parent_bam.header

        self._outqu = outqu
        self._inqu = inqu

        self._task_size = task_size
        self._temp_dir = constants.temp_dir
        self._constants = constants 
        
        self._dealt = 0

    def run(self):

        bamfile = pysam.AlignmentFile(self._parent_path,"rb")
        iterator = bamfile.fetch(until_eof=True)
        next_read = iterator.next

        generate_results = self.__generate_results__

        while True:
            try:
                package = self._inqu.get(False)
            
                if type(package) == DestroyPackage:
                    bamfile.close()
                    del iterator
                    del bamfile
                    break

                seek = package                
                bamfile.seek(seek)
                time.sleep(.01)

                results = generate_results(iterator)
                
                self._dealt += 1
                time.sleep(0.005)

                self._outqu.put(Package(results=results))

            except Queue2.Empty:
                time.sleep(5)
            except StopIteration:
                results = self.__handle_stop_iteration__(seek,bamfile,iterator)
                self._dealt += 1 #Copy paste code here.
                time.sleep(0.005) #And here
                self.__post_run_routine__()
                self._outqu.put(Package(results=results))
        return

    def __generate_results__(self,iterator,**kwargs):
        self.__pre_run_routine__(iterator)
        self.__process_task_set__(iterator)
        results = self.__get_results__()
        results["total"] = self._task_size
        self.__post_run_routine__()
        return results

    def __handle_stop_iteration__(self,seek,bamfile,iterator):
        results = self.__get_results__()
         
        cdef int count = 0 
        bamfile.seek(seek)
        try: #Count the amount of reads until end of file
            for read in iterator:
                count += 1
        except StopIteration:
            pass

        results["total"] = count
        return results

    def __get_temp_object__(self,path):
        return pysam.AlignmentFile(path,"wb",header=self._parent_bam.header)

    def __get_temp_path__(self,identity):
        file_name = "%s_%d_%d_%s" %\
            (identity,self.pid,self._dealt, os.path.split(self._parent_bam.filename)[1])
        return os.path.join(self._temp_dir,file_name)

    @abstractmethod
    def __pre_run_routine__(self,iterator,**kwargs):
        pass

    @abstractmethod
    def __process_task_set__(self,iterator,**kwargs):
        pass

    @abstractmethod
    def __get_results__(self):
        pass

    @abstractmethod
    def __post_run_routine__(self,**kwargs):
        pass

#The FileReader iterates over a BAM, subsets reads and
#then starts a parbam.Task process on the subsetted reads
class FileReader(Process):
    def __init__(self,str input_path,int proc_id,object outqu,
                 int task_n,object constants,object Task,
                 object pause_qu):
        
        super(FileReader,self).__init__()

        self._input_path = input_path
        self._proc_id = proc_id
        
        self._outqu = outqu
        self._pause_qu = pause_qu

        self._Task = Task
        self._task_n = task_n

        self._constants = constants
        self._task_size = constants.task_size
        self._reader_n = constants.reader_n
        self._temp_dir = constants.temp_dir
        self._debug = constants.debug

    #Find data pertaining to assocd and all reads 
    #and divide pertaining to the chromosome that it is aligned to
    def run(self):

        parent_bam_mem_obj = ParentAlignmentFile(self._input_path)
        parent_bam = pysam.AlignmentFile(self._input_path,"rb")
        parent_iter = self.__get_parent_iter__(parent_bam)
        parent_generator = self.__bam_generator__(parent_iter)
        wait_for_pause = self.__wait_for_pause__
        if self._debug:
            parent_generator = self.__debug_generator__(parent_iter)

        task_qu = Queue()

        tasks = [self._Task(parent_bam=parent_bam_mem_obj,
                      inqu=task_qu,
                      outqu=self._outqu,
                      task_size=self._task_size,
                      constants=self._constants) \
                        for i in xrange(self._task_n)]

        for task in tasks:
            task.start()

        for i,command in enumerate(parent_generator):
            #This pause adds stability to parabam
            #I think this is because it spaces calls to disk
            time.sleep(.0001) 
            wait_for_pause()
            task_qu.put(parent_bam.tell())

        for n in xrange(self._task_n+1):
            task_qu.put(DestroyPackage())

        time.sleep(3)
        parent_bam.close()
        task_qu.close()
        return

    def __get_parent_iter__(self,parent_bam):
        if not self._constants.fetch_region:
            return parent_bam.fetch(until_eof=True)
        else:
            return parent_bam.fetch(region=self._constants.fetch_region)

    def __bam_generator__(self,parent_iter):
        cdef int proc_id = self._proc_id
        cdef int reader_n = self._reader_n
        cdef int iterations = 0
        cdef int task_size = self._task_size

        while True:
            try:
                if iterations % reader_n == proc_id:
                    yield True
                for x in xrange(task_size):
                    parent_iter.next()
                iterations += 1
            except StopIteration:
                break
        return

    def __debug_generator__(self,parent_iter):
        cdef int proc_id = self._proc_id
        cdef int reader_n = self._reader_n
        cdef int iterations = 0
        cdef int task_size = self._task_size

        while True:            
            try:
                if iterations % reader_n == proc_id:
                    yield True
                for x in xrange(task_size):
                    parent_iter.next()
                iterations += 1

                if iterations == 25:
                    break

            except StopIteration:
                break
        return

    def __send_ack__(self,qu):
        qu.put(2)
        print "ACK SENT"
        time.sleep(1)

    def __query_pause_qu__(self,bypass,pause_qu):
        try:
            if bypass:
                pause = True
            else:
                pause = pause_qu.get(False)
                self.__send_ack__(pause_qu)

            if pause:
                while True:
                    try:
                        pause = pause_qu.get(False)
                        if not pause:
                            self.__send_ack__(pause_qu)#Acknowledge
                            break
                    except Queue2.Empty:
                        time.sleep(.5)
        except Queue2.Empty:
            pass

        #code is to fix a race condition when a
        # pause can be missed due to 
        #waiting on a previous pause
        last = False 
        while True:
            try:
                pause = pause_qu.get(False)
                last = pause
            except Queue2.Empty:
                break
        return last

    def __wait_for_pause__(self):
        pause_qu = self._pause_qu
        pause = self.__query_pause_qu__(False,pause_qu)
        if pause:
            while True:
                pause = self.__query_pause_qu__(True,pause_qu)
                if not pause:
                    break
        
class Leviathon(object):
    #Leviathon takes objects of file_readers and handlers and
    #chains them together.
    def __init__(self,object constants,dict handler_bundle,
                 list handler_order,list queue_names,
                 int update,object Task):

        self._constants = constants
        self._handler_bundle  = handler_bundle
        self._handler_order = handler_order
        self._update = update
        self._queue_names = queue_names
        self._Task = Task
    
    def run(self,input_path,output_paths):
        parent = ParentAlignmentFile(input_path)

        default_qus = self.__create_queues__(self._queue_names) 
        pause_qus = self.__create_pause_qus__(self._constants.reader_n)


        handlers_objects,handler_inqus = self.__create_handlers__(self._handler_order,
                                                          self._handler_bundle,self._constants,
                                                          default_qus,parent,output_paths,pause_qus)
        handlers = self.__get_handlers__(handlers_objects)

        task_n = self.__get_task_n__(self._constants,handlers)
        file_reader_bundles = self.__get_file_reader_bundles__(default_qus,parent,
                                                               task_n,self._constants,pause_qus)
        file_readers = self.__get_file_readers__(file_reader_bundles)

        #Start file_readers
        for file_reader in file_readers:
            file_reader.start()
            #time.sleep(2)

        #Start handlers:
        for handler in handlers:
            handler.start()

        #Wait for file_readers to finish
        for file_reader in file_readers:
            file_reader.join()

        #Inform handlers that processing has finished
        for handler,queue in izip(handlers,handler_inqus):
            queue.put(EndProcPackage())

        #Destory handlers
        for handler,queue in izip(handlers,handler_inqus):
            queue.put(DestroyPackage())
            handler.join()

        del default_qus
        del file_reader_bundles
        del file_readers
        del pause_qus
        del handler_inqus
        del handlers_objects
        del handlers

        gc.collect()

    def __create_pause_qus__(self,reader_n):
        return [Queue() for i in xrange(reader_n)]

    def __get_task_n__(self,constants,handlers):
        task_n = (constants.total_procs - constants.reader_n) / constants.reader_n
        if task_n > 0:
            return task_n
        else:
            return 1

    def __get_file_readers__(self,file_reader_bundles):
        file_readers = []
        for bundle in file_reader_bundles: 
            file_reader = FileReader(**bundle)
            file_readers.append(file_reader)
        return file_readers

    def __proc_id_generator__(self,reader_n):
        for i in reversed(xrange(reader_n)):
            yield i

    def __get_file_reader_bundles__(self,default_qus,parent,task_n,constants,pause_qus):
        bundles = []

        for proc_id,pause in izip(self.__proc_id_generator__(constants.reader_n),pause_qus):
            current_bundle = {"input_path" :parent.filename,
                              "proc_id" :proc_id,
                              "task_n" :task_n,
                              "outqu" :default_qus["main"],
                              "constants" :constants,
                              "Task" :self._Task,
                              "pause_qu" :pause}

            bundles.append(current_bundle)

        return bundles
 
    def __create_queues__(self,queue_names):
        queues = {}
        for name in queue_names:
            queues[name] = Queue()
        return queues

    def __create_handlers__(self,handler_order,handler_bundle,constants,
                            queues,parent_bam,output_paths,pause_qus):
        handlers = []
        handler_inqus = []

        for handler_class in handler_order:
            handler_args = dict(handler_bundle[handler_class])

            handler_args["parent_bam"] = parent_bam
            handler_args["output_paths"] = output_paths
            handler_args["constants"] = constants

            #replace placeholder with queues
            handler_args["pause_qus"] = pause_qus
            handler_args["inqu"] = queues[handler_args["inqu"]]
            handler_args["out_qu_dict"] = dict(\
                    [(name,queues[name]) for name in handler_args["out_qu_dict"] ])

            handler_inqus.append(handler_args["inqu"])
            handlers.append(handler_class(**handler_args))

        return handlers,handler_inqus

    def __get_handlers__(self,handlers):
        handler_processes = []
        for handler in handlers:
            hpr = Process(target=handler.listen,args=(self._update,))
            handler_processes.append(hpr)
        return handler_processes

#Provides a conveinant way for providing an Interface to parabam
#programs. Includes default command_args and framework for
#command-line and programatic invocation. 
class Interface(object):

    __metaclass__ = ABCMeta

    def __init__(self,temp_dir):
        self._temp_dir = temp_dir

    def __introduce__(self,name):
        intro =  "%s has started. Start Time: " % (name,)\
            + datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
        underline = "-" * len(intro)
        print intro
        print underline

    def __goodbye__(self,name):
        print "%s has finished. End Time: " % (name,)\
            + datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')

    def __sort_and_index__(self,fnm,verbose=False,tempDir=".",name=False):
        if not os.path.exists(fnm+".bai"):
            if verbose: print "%s is not indexed. Sorting and creating index file." % (fnm,)
            tempsort_path = self.__get_unique_temp_path__("SORT",temp_dir=self._temp_dir)
            optStr = "-nf" if name else "-f"
            pysam.sort(optStr,fnm,tempsort_path)
            os.remove(fnm)
            os.rename(tempsort_path,fnm)
            if not name: pysam.index(fnm)

    def __get_unique_temp_path__(self,temp_type,temp_dir="."):
        #TODO: Duplicated from parabam.interface.merger. Find a way to reformat code
        #to remove this duplication
        return "%s/%sTEMP%d.bam" % (temp_dir,temp_type,int(time.time()),)

    def __get_group__(self,bams,names=[],multi=1):
        if multi == 0:
            multi = 1
        if names == []:
            names = [self.__get_basename__(path) for path in bams]
        for i in xrange(0,len(bams),multi):
            yield (bams[i:i+multi],names[i:i+multi])

    def __get_basename__(self,path):
        base,ext = os.path.splitext(os.path.basename(path))
        return base

    def default_parser(self):

        parser = ParabamParser(conflict_handler='resolve',
                    formatter_class=argparse.RawTextHelpFormatter)

        parser.add_argument('-p',type=int,nargs='?',default=4
            ,help="The maximum amount of processes you wish parabam to use. This should"\
                  "be less than or equal to the amount of processor cores in your machine.")
        parser.add_argument('-s',type=int,nargs='?',default=250000
            ,help="The amount of reads considered by each distributed task.")
        parser.add_argument('-f',type=int,metavar="READERS",nargs='?',default=2
            ,help="The amount of open connections to the file being read. Conventional hard drives\
            perform best with the default of 2\n")
        return parser

    @abstractmethod
    def run_cmd(self,parser):
        #This is usualy just a function that
        #takes an argparse parser and turns 
        #passes the functions to the run function
        pass

    @abstractmethod
    def run(self):
        pass

    @abstractmethod
    def get_parser(self):
        pass

class ParabamParser(argparse.ArgumentParser):
    def error(self, message):
        self.print_help()
        sys.stderr.write('\nerror: %s\n' % message)
        sys.exit(2)

class Constants(object):
    
    def __init__(self,temp_dir,verbose,task_size,total_procs,fetch_region,**kwargs):
        #TODO: Update with real required constants values
        self.temp_dir = temp_dir
        self.verbose = verbose
        self.task_size = task_size
        self.total_procs = total_procs
        self.fetch_region = fetch_region

        for key, val in kwargs.items():
            setattr(self,key,val)

    def add(self,key,val):
        setattr(self,key,val)

class Package(object):
    def __init__(self,results):
        self.results = results

class DestroyPackage(Package):
    def __init__(self):
        super(DestroyPackage,self).__init__(results={})
        self.destroy = True

class EndProcPackage(Package):
    def __init__(self):
        super(EndProcPackage,self).__init__(results={})
        self.destroy = False

class ParentAlignmentFile(object):
    
    def __init__(self,path):
        has_index = os.path.exists(os.path.join("%s%s" % (path,".bai")))

        parent = pysam.AlignmentFile(path,"rb")
        self.filename = parent.filename
        self.references = parent.references
        self.header = parent.header
        self.lengths = parent.lengths

        if has_index:
            self.nocoordinate = parent.nocoordinate
            self.nreferences = parent.nreferences
            self.unmapped = parent.unmapped
        else:
            self.mapped = 0
            self.nocoordinate = 0
            self.nreferences = 0
            self.unmapped = 0

        parent.close()

    def getrname(self,tid):
        return self.references[tid]

    def gettid(self,reference):
        for i,ref in enumerate(self.references):
            if reference == ref:
                return i
        return -1


#And they all lived happily ever after...



import pysam
import parabam
import time
import sys
import os
import gc
import shutil

from itertools import izip
from multiprocessing import Queue

class SubsetCore(object):

    def __init__(self,object constants,**kwargs):
        self.constants = constants

    def __pre_run_routine__(self,iterator,**kwargs):
        super(SubsetCore,self).__pre_run_routine__(iterator)
        self._user_subsets = self.constants.user_subsets
        self._counts = {}
        self._temp_paths = {}
        self._temp_objects = {}

        for subset in self._user_subsets:
            self._temp_paths[subset] = self.__get_temp_path__(subset)
            self._temp_objects[subset] = self.__get_temp_object__(self._temp_paths[subset])
            self._counts[subset] = 0

    def __get_results__(self):
        results = {}
        results["temp_paths"] = self._temp_paths
        results["counts"] = self._counts
        results["system"] = self._system

        return results

    def __write_to_subset_bam__(self,subset_type,read):
        self._counts[subset_type] += 1
        self._temp_objects[subset_type].write(read)

    def __post_run_routine__(self,**kwargs):
        super(SubsetCore,self).__post_run_routine__()
        for subset,fileobj in self._temp_objects.items():
            fileobj.close()


class Task(SubsetCore,parabam.command.Task):

    def __init__(self,parent_bam,inqu,outqu,statusqu,task_size,constants,**kwargs):
        
        parabam.command.Task.__init__(self,parent_bam=parent_bam,
                                    inqu=inqu,
                                    outqu=outqu,
                                    statusqu=statusqu,
                                    task_size=task_size,
                                    constants=constants)
        SubsetCore.__init__(self,constants)

    def __handle_engine_output__(self,engine_output,read):             
        if type(engine_output) == bool:
            if engine_output:
                self.__write_to_subset_bam__(self._constants.user_subsets[0],read)          
        elif type(engine_output) == list:
            for subset,cur_read in engine_output:
                self.__write_to_subset_bam__(subset,cur_read)
        elif type(engine_output) == None:
            pass

class PairTask(SubsetCore,parabam.command.PairTask):
    def __init__(self,parent_bam,inqu,outqu,statusqu,task_size,constants):
        
        parabam.command.PairTask.__init__(self,parent_bam=parent_bam,
                                                inqu=inqu,
                                                outqu=outqu,
                                                statusqu=statusqu,
                                                task_size=task_size,
                                                constants=constants)
        SubsetCore.__init__(self,constants)

    def __handle_engine_output__(self,engine_output,read):
        for subset,cur_read in engine_output:
            self.__write_to_subset_bam__(subset,cur_read)

class Handler(parabam.command.Handler):

    def __init__(self,object parent_bam, object output_paths,object inqu,
                object constants,object pause_qus,dict out_qu_dict):
        
        super(Handler,self).__init__(parent_bam = parent_bam,output_paths = output_paths,
                                     inqu=inqu,constants=constants,pause_qus=pause_qus,
                                     out_qu_dict=out_qu_dict)

        self._user_subsets = constants.user_subsets
        for subset in self._user_subsets:
            self._stage_stores[subset] = []

    def __new_package_action__(self,new_package):
        super(Handler,self).__new_package_action__(new_package)
        results = new_package.results
        for subset in self._user_subsets:
            self._stage_stores[subset].append((results["counts"][subset],
                                               results["temp_paths"][subset],))

    def __periodic_action__(self,iterations):
        super(Handler,self).__periodic_action__(iterations)

        for subset in self._user_subsets:
            if self.__test_stage_store__(subset):
                self.__add_merge_task__(results=self._stage_stores[subset],subset_type=subset)
                self.__update_stage_store__(subset)

    def __handler_exit__(self, **kwargs):
        super(Handler,self).__handler_exit__()
        #Kill the handlers handler

        for subset in self._user_subsets:
            #merge task 
            self.__add_merge_task__(self._stage_stores[subset],subset)
        self.__write_counts_csv__()
    
    def __add_merge_task__(self,results,subset_type):
        res = parabam.merger.MergePackage(results=results,subset_type=subset_type)
        self._out_qu_dict["merge"].put(res)

    def __write_counts_csv__(self):
        if self._constants.output_counts:
            count_path = "./subset_counts_csv_%d.csv" % (time.time(),)
            self.__standard_output__("[Status] Outputting counts to following file: %s" % (count_path,))

            with open(count_path, "w") as count_file:
                key_order,header = self.__get_count_header__()
                count_file.write(header)
                for line in self.__generate_count_line__(key_order):
                    count_file.write(line)

    def __generate_count_line__(self,key_order):
        for source,stats in self._stats.items():
            line = ""
            line += "%s" % (source,)
            for key in key_order:
                line += ",%s" % (stats[key],)
            line += "\n"
            yield line

    def __get_count_header__(self):
        header = "Sample"
        keys = []
        for source, stats in self._stats.items():
            for key, val in stats.items():
                header += ","
                header += key
                keys.append(key)
            break
        header += "\n"
        return keys, header

class Interface(parabam.command.Interface):
    """The interface to parabam subset.
    Users will primarily make use of the ``run`` function."""

    def __init__(self,temp_dir):
        super(Interface,self).__init__(temp_dir)

    def run_cmd(self,parser):
        cmd_args = parser.parse_args()

        module,user_engine,user_constants = self.__get_module_and_vitals__(cmd_args.instruc)

        if hasattr(module,"get_subset_types"):
            user_subsets = module.get_subset_types()
        else:
            user_subsets = ["subset"]

        self.run(
            input_bams=cmd_args.input,
            total_procs = cmd_args.p,
            task_size = cmd_args.s,
            verbose= cmd_args.v,
            user_subsets= user_subsets,
            user_constants = user_constants,
            user_engine = user_engine,
            fetch_region = cmd_args.region,
            reader_n = cmd_args.f,
            pair_process=cmd_args.pair,
            include_duplicates=cmd_args.d,
            debug = cmd_args.debug,
            ensure_unique_output=cmd_args.u,
            output_counts=cmd_args.counts,
            announce = True
            )
    
    def run(self,input_bams,total_procs,task_size,user_constants,user_engine,
            user_subsets,reader_n = 2,fetch_region=None,
            keep_in_temp=False,verbose=0,pair_process=False,
            include_duplicates=True,debug=False,
            ensure_unique_output=False,output_counts=False,announce=False):
        ''' Docstring! '''

        args = dict(locals())
        del args["self"]

        if not verbose:
            announce = False
        self.__introduce__("parabam subset",announce)

        results = super(Interface,self).run(**args)
        
        self.__goodbye__("parabam subset",announce)
        return results

    def __get_queue_names__(self,pair_process,**kwargs):
        queues = ["merge","main"]
        return queues

    def __get_destroy_handler_order__(self):
        return [Handler,parabam.merger.Handler]

    def __get_handler_bundle__(self,**kwargs):
        handler_bundle = {}

        #queues transformed by leviathon
        handler_bundle[Handler] = {"inqu":"main", 
                                   "out_qu_dict":["merge"]}

        handler_bundle[parabam.merger.Handler] = {"inqu":"merge",
                                                  "out_qu_dict":[]}
        return handler_bundle

    def __get_output_paths__(self,input_path,user_subsets,ensure_unique_output,**kwargs):
        output_paths = {input_path:{}}
        for salt,subset in enumerate(user_subsets):
            output_paths[input_path][subset] = self.__get_path__(input_path,
                                                                 subset,
                                                                 ensure_unique_output,
                                                                 salt)
        return output_paths

    def __get_path__(self,input_path,subset,ensure_unique_output,salt):
        head,ext = os.path.splitext(input_path)
        head,tail = os.path.split(head)
        unique = ""

        if ensure_unique_output:
            unique = "_%d%d" % (time.time(),salt)

        return os.path.join(".",self._temp_dir, "%s_%s%s%s" % (tail,subset,unique,ext))
            
    def __get_task_class__(self,pair_process,**kwargs):
        if pair_process:
            return PairTask
        else:
            return Task

    def __get_extra_const_args__(self,**kwargs):
        args = {}
        args["user_subsets"] = kwargs["user_subsets"]
        args["output_counts"] = kwargs["output_counts"]
        return args
        
    def get_parser(self):
        parser = self.default_parser()

        parser.add_argument('-u',action="store_true",default=False,
            help="The files will be appended with a unique identifier. In the format <input_name>_<unique_id>_<subset_type>.bam")
        parser.add_argument('--counts',action="store_true",default=False,
            help="The amount of reads in each subset will be output as a .CSV alongside the .BAM file.")

        return parser 

#...happily ever after



#Once upon a time
import pysam
import time
import sys
import os
import gc
import parabam
import random

import Queue as Queue2
import numpy as np

from parabam.core import DestroyPackage
from itertools import izip
from collections import Counter,namedtuple
from multiprocessing import Queue,Process

class ChaserResults(parabam.core.Package):
    def __init__(self,results,chaser_type):
        super(ChaserResults,self).__init__(results)
        self.chaser_type=chaser_type

class PrimaryResults(ChaserResults):
    def __init__(self,results,chaser_type):
        super(PrimaryResults,self).__init__(results,chaser_type)

class MatchMakerResults(ChaserResults):
    def __init__(self,loner_type,results,chaser_type,rescued,level):
        super(MatchMakerResults,self).__init__(results,chaser_type)
        self.rescued = rescued
        self.loner_type = loner_type
        self.level = level

class JobPackage(object):
    def __init__(self,paths,loner_type,level):
        self.paths = paths
        self.loner_type = loner_type 
        self.level = level

class Handler(parabam.core.Handler):

    def __init__(self,object parent_bam, object output_paths,
                      object inqu,object constants,object pause_qus,
                      dict out_qu_dict,object Task):

        super(Handler,self).__init__(parent_bam = parent_bam,output_paths = output_paths,
                             inqu=inqu,constants=constants,pause_qus=pause_qus,
                             out_qu_dict=out_qu_dict,report = False)
        
        class EngineTask(Task):
            def __init__(self,parent_bam,constants):
                super(EngineTask,self).__init__(parent_bam=parent_bam,
                                                    inqu=None,
                                                    outqu=None,
                                                    statusqu=None,
                                                    task_size=0,
                                                    constants=constants)

            def handle_pair_dict(self,pairs,unique):
                self.unique = unique
                results = self.__generate_results__(pairs)
                results["total"] = 0
                self._dealt += 1
                time.sleep(.01)
                return results

            def __process_task_set__(self,iterator):
                engine = self._engine
                parent_bam = self._parent_bam
                handle_output = self.__handle_engine_output__
                user_constants = self._user_constants

                for read_name,pair in iterator.items():
                    engine_output = engine(pair,user_constants,parent_bam)
                    handle_output(engine_output,pair)

            def __get_temp_path__(self,identity):
                file_name = "%s_%d_%d_%s" %\
                    (identity,self.unique,self._dealt, os.path.split(self._parent_bam.filename)[1])
                return os.path.join(self._temp_dir,file_name)

        self._loner_pyramid = self.__instalise_loner_pyramid__()
        self._pyramid_idle_counts = Counter()
        self._loner_purgatory = {}

        self._periodic_interval = 5
        self._total_loners = 0
        self._rescued = Counter()
        self._prev_rescued = 0
        self._stale_count = 0

        self._child_pack = {"parent_bam":parent_bam,
                            "queue":out_qu_dict["main"],
                            "chrom_bins":self.__get_chrom_bins__(parent_bam) ,
                            "chaser_task":EngineTask(parent_bam,self._constants)}

        self._chaser_qu = Queue()
        self._chaser_tasks = self.__create_chaser_tasks__(self._constants.total_procs)
        self._pending_jobs = 0
        self._max_jobs = (self._constants.total_procs * 3)
        self._file_readers_paused = False

        self._post_destroy_count = 0
        self._post_destroy_thresh = 50

        self._primary_store = self.__instalise_primary_store__()

    def __create_chaser_tasks__(self,total_tasks):
        tasks = []
        for i in xrange(total_tasks):
            task = ChaserTask(constants=self._constants,
                              inqu=self._chaser_qu,
                              outqu=self._inqu,
                              max_reads=7000000,
                              child_pack=self._child_pack)
            task.start()
            tasks.append(task)
        return tasks

    def __get_chrom_bins__(self,parent_bam):
        references = parent_bam.references
        lengths = parent_bam.lengths

        chrom_bins = {}
        chrom_info = zip(xrange(len(references)),references,lengths)
        chrom_info = sorted(chrom_info, key=lambda tup: tup[2])
        group = []
        size = 0
        uid = 0
        for i,name,length in chrom_info:
           group.append( (i,name,length,) )
           size += length

           if size > 150000000:
                for i,name,length in group:
                    chrom_bins[i] = "b%d" % (uid,)
                group = []
                size = 0
                uid += 1
        return chrom_bins
           
    def __instalise_primary_store__(self):
        return []

    def __instalise_loner_pyramid__(self):
        return {}

    def __update_tasks__(self,active_tasks):
        terminated_procs = []
        for pr in active_tasks:
            if not pr.is_alive():
                pr.terminate()
                terminated_procs.append(pr)
        
        for pr in terminated_procs:
            active_tasks.remove(pr)

        if len(terminated_procs) > 0:
            #Only invoked the GC if there is the possibility of
            #collecting any memory.
            del terminated_procs
            gc.collect()

        return len(active_tasks)

    def __new_package_action__(self,new_package,**kwargs):        
        chaser_type = new_package.chaser_type
        if chaser_type == "origin":
            self.__handle_origin_task__(new_package)
        elif chaser_type == "match_maker":
            self.__handle_match_maker__(new_package)
            self._pending_jobs -= 1
        elif chaser_type == "primary":
            self.__handle_primary_task__(new_package)
            self._pending_jobs -= 1
        else:
            print "Error!"

    def __handle_origin_task__(self,new_package):
        for loner_count,path in new_package.results:
            if loner_count == 0:
                os.remove(path)
            else:
                self._total_loners += loner_count
                self._primary_store.append(path) 

    def __handle_match_maker__(self,new_package):
        loner_type = new_package.loner_type
        level = new_package.level

        #this number is by pairs to get read num we times 2
        self._rescued["total"] += (new_package.rescued*2)    

        if new_package.results: #This ensures that there are leftover loners
            self._pyramid_idle_counts[loner_type] = 0
            
            try:
                pyramid = self._loner_pyramid[loner_type]
            except KeyError:
                self._loner_pyramid[loner_type] = [[]]
                pyramid = self._loner_pyramid[loner_type]

            if len(pyramid) >= level:
                pyramid.append([])

            for loner_level,loner_path in new_package.results:
                pyramid[loner_level].append(loner_path)

        if self._destroy:
            self._stale_count += 1

    def __handle_primary_task__(self,new_package):
        for match_package in new_package.results:
            self.__handle_match_maker__(match_package)

    def __test_primary_tasks__(self,required_paths=10):
        task_sent = False
        if len(self._primary_store) >= required_paths:
            self.__start_primary_task__()
            task_sent = True

        if task_sent:
            self._primary_store = []
            gc.collect()
            time.sleep(1)
          
    def __start_primary_task__(self):        
        self.__start_job__(self._primary_store,"Primary",-1)

    def __start_matchmaker_task__(self,paths,loner_type,level):
        self.__start_job__(paths,loner_type,level)

    def __start_job__(self,paths,loner_type,level):
        job = JobPackage(paths,loner_type,level)
        self._pending_jobs += 1
        self._chaser_qu.put(job)

    def __pause_monitor__(self):
        if not self._destroy:
            max_jobs = self._max_jobs
            if self._pending_jobs >= max_jobs and not self._file_readers_paused:
                for qu in self._pause_qus:
                    qu.put(1) #pause
                    self.__wait_for_ack__(qu)
                self._file_readers_paused = True

            elif self._pending_jobs < max_jobs and self._file_readers_paused:
                for qu in self._pause_qus:
                    qu.put(0) #unpause
                    self.__wait_for_ack__(qu)
                self._file_readers_paused = False

    def __wait_for_ack__(self,qu):
        count = 0
        while True:
            if count > 10:
                return
            try:
                ack = qu.get(False)
                if ack == 2:
                    return
                else:
                    qu.put(ack)
                    time.sleep(1)
            except Queue2.Empty:
                time.sleep(1)
            count += 1

    def __periodic_action__(self,iterations):

        self.__pause_monitor__()
        chaser_debug = False

        if not self._destroy:
            idle_threshold = 500
            required_paths = 10
        else:
            idle_threshold = 75
            required_paths = 1

        pyramid_idle_counts = self._pyramid_idle_counts 
        empty = True

        for loner_type,pyramid in self._loner_pyramid.items():
            for i,sub_pyramid in enumerate(reversed(pyramid)):
                level = len(pyramid) - (i+1)
                paths = self.__get_paths__(loner_type,sub_pyramid,level)
                if len(paths) > 0:
                    empty = False
                    self.__start_matchmaker_task__(paths,loner_type,level)
                    pyramid_idle_counts[loner_type] = 0
                    for path in paths:#remove sent paths
                        sub_pyramid.remove(path)
                    break
                if len(sub_pyramid) > 0:
                    empty = False
                    pyramid_idle_counts[loner_type] += 1
                del paths

            if pyramid_idle_counts[loner_type] >= idle_threshold:
                idle_success,idle_paths,idle_levels = self.__idle_routine__(pyramid,loner_type)

                #delete after idle success or when processing has finished and idle fails
                if self.__clear_subpyramid__(idle_success, self._destroy,self._pending_jobs):
                    pyramid_idle_counts[loner_type]=0
                    for idle_level,idle_path in izip(idle_levels,idle_paths):
                        if not idle_success:
                            try:
                                self._loner_purgatory[loner_type].append(idle_path)
                            except KeyError:
                                self._loner_purgatory[loner_type] = [idle_path]
                            pyramid[idle_level].remove(idle_path)
                        else:
                            pyramid[idle_level].remove(idle_path)

                del idle_paths,idle_levels

        self.__test_primary_tasks__(required_paths=required_paths)

        if self._destroy:
            self.__post_destroy_report__()
            if not self._rescued["total"] == self._prev_rescued:
                self._stale_count = 0
            self._prev_rescued = self._rescued["total"]
            saved = self.__save_purgatory_loners__()

            if (empty and self._pending_jobs == 0) or self._stale_count > 100:
                finished = True
                if not empty: #essentialy `if stale count`
                    self.__tidy_pyramid__()
                else: 
                    if saved > 0:
                        finished = False
                    if not self.__is_queue_empty__():
                        finished = False
                                       
                if finished:
                    self._finished = True

        if iterations % 20 == 0:
            gc.collect()

        if chaser_debug and iterations % 30 == 0:
            sys.stdout.write("\r %d/%d=%.4f %.2fGB | Proc:%d Des:%d Emp:%d Purg:%d Stale:%d Jobs:%d " %\
                (self._rescued["total"],
                self._total_loners,
                float(self._rescued["total"]+1)/(self._total_loners+1),
                float((self._total_loners-self._rescued["total"]) * 130) / (10**9),
                self._processing,
                self._destroy,
                empty,
                len(self._loner_purgatory),
                self._stale_count,
                self._pending_jobs))
            sys.stdout.flush()

    def __is_queue_empty__(self):
        try:
            pack = self._inqu.get(False,60)
            self._inqu.put(pack)
            return False
        except Queue2.Empty:
            return True

    def __save_purgatory_loners__(self):
        saved = 0
        saved_keys = []
        for loner_type,paths in self._loner_purgatory.items():
            if len(paths) > 1:
                saved += 1
                saved_keys.append(loner_type)
                self._pyramid_idle_counts[loner_type] = 100
                for path in paths:
                    self._loner_pyramid[loner_type][0].append(path)
        for key in saved_keys:
            del self._loner_purgatory[key]
        gc.collect()
        return saved

    def __get_paths__(self,loner_type,sub_pyramid,level):
        if not self._destroy and "UM" in loner_type:
            return []
        task_size = self.__get_task_size__(level,loner_type)
        if len(sub_pyramid) >= task_size:
            if "XX" in loner_type:
                return sub_pyramid[:task_size]
            else:
                return random.sample(sub_pyramid,task_size)
        return []
    
    def __tidy_pyramid__(self):
        for loner_type,pyramid in self._loner_pyramid.items():
            for sub_pyramid in pyramid:
                for x in xrange(len(sub_pyramid)):
                    os.remove(sub_pyramid.pop())
        gc.collect()
            
    def __clear_subpyramid__(self,success,destroy,running):
        if success:
            #If the idle task was sent, clear pyramid
            return True
        elif destroy and not success:
            if running == 0:
                #if main process has finished and no tasks run
                #clear the pyramid
                return True
            else:
                #Tasks still running. Don't clear
                return False 
        else:
            #Idle task not sent, don't clear pyramid
            return False

    def __idle_routine__(self,pyramid,loner_type):
        if "UM" in loner_type and not self._destroy:
            #Don't conduct idle on UMs because we don't expect to see pairs
            #until after processing has finished
            return True,[],[]

        all_paths = []
        for level,sub_pyramid in enumerate(pyramid):
            all_paths.extend(zip([level]*len(sub_pyramid),sub_pyramid))

        success = False
        idle_paths = []
        idle_levels = []
        if len(all_paths) > 1:
            path_n = random.randint(0,len(all_paths))
            if path_n < 2:
                #This gives 3 chances to roll a 2
                #2 paths means a depth heavy search
                path_n = 2

            idle_tuples = random.sample(all_paths,path_n)
            idle_levels,idle_paths = zip( *idle_tuples )

            self.__start_matchmaker_task__(idle_paths,loner_type,max(idle_levels))
            success = True
            del all_paths,idle_tuples
        elif len(all_paths) == 1:
            idle_levels,idle_paths = zip(*all_paths)

        return success,idle_paths,idle_levels

    def __get_task_size__(self,level,loner_type):
        if level % 2 == 0:
            task_size = 2
        else:
            task_size = 3
        return task_size

    def __handler_exit__(self,**kwargs):
        for i in xrange(len(self._chaser_tasks)+1):
           self._chaser_qu.put(DestroyPackage())
        time.sleep(2)
        self._chaser_qu.close()
        del self._chaser_qu
        del self._chaser_tasks
        
        if self._constants.verbose:
            if self._total_loners - self._rescued["total"] == 0:
                if self._post_destroy_count >= self._post_destroy_thresh:
                    sys.stdout.write("\r[Status] Read pairing still in progress: 100.000%% complete")
                    sys.stdout.flush()
                self.__standard_output__("\n[Status] All reads succesfully paired") 
            else:
                self.__standard_output__("\n[Status] Couldn't find pairs for %d reads" %\
                    (self._total_loners - self._rescued["total"],))
        for qu in self._pause_qus:
            qu.close()

    def __post_destroy_report__(self):
        if self._constants.verbose:
            if self._post_destroy_count >= self._post_destroy_thresh:
                if self._post_destroy_count == self._post_destroy_thresh:
                    sys.stdout.write("\n")
                    sys.stdout.flush()
                if self._post_destroy_count % 5 == 0:
                    sys.stdout.write("\r[Status] Read pairing still in progress: %.3f%% complete  " %\
                                ((float(self._rescued["total"]+1) / (self._total_loners+1))*100,))
                    sys.stdout.flush()
            self._post_destroy_count += 1

class ChaserTask(Process):
    def __init__(self,object constants,object inqu, object outqu, 
                      int max_reads, dict child_pack):
        super(ChaserTask,self).__init__()
        self._constants = constants
        self._inqu = inqu
        self._outqu = outqu
        self._child_pack = child_pack
        self._max_reads = max_reads

    def run(self):
        primary_handler = PrimaryHandler(constants = self._constants,
                                         max_reads = self._max_reads,
                                         child_pack = self._child_pack,
                                         pid = self.pid)

        match_handler = MatchMakerHandler(constants = self._constants,
                                         max_reads = self._max_reads,
                                         child_pack = self._child_pack,
                                         pid = self.pid)

        while True:
            try:
                package = self._inqu.get(False)
                if type(package) == JobPackage:
                    if package.level == -1:
                        results = primary_handler.run(package.paths)
                    else:
                        results = match_handler.run(package.paths,
                                          package.loner_type,
                                          package.level)
                    self._outqu.put(results)

                elif type(package) == DestroyPackage:
                    break
                
                time.sleep(0.005)

            except Queue2.Empty:
                time.sleep(5)
        return

class ChaserHandler(object):

    def __init__(self,object constants,int max_reads, dict child_pack,int pid):
        self._constants = constants
        self._max_reads = max_reads
        self._child_pack = child_pack
        self._parent_bam = self._child_pack["parent_bam"]
        self._header = self._parent_bam.header
        self._record = 0
        self.pid = pid

    def __get_loner_temp_path__(self,loner_type,level=0,unique=""):
        self._record += 1
        return "%s/%s_%d%d_%s%s.bam" %\
            (self._constants.temp_dir,self.pid,self._record,level,loner_type,unique)

class PrimaryHandler(ChaserHandler):

    def __init__(self,object constants, int max_reads, dict child_pack,int pid):
        super(PrimaryHandler,self).__init__(constants,max_reads,child_pack,pid)
        self.match_handler = MatchMakerHandler(constants,max_reads,child_pack,pid)
        
    def run(self,unsorted_paths):
        #speedup
        write_loner = self.__write_loner__
        #speedup

        loner_holder = self.__get_loner_holder__()
        loner_file_counts = Counter()

        loner_total = 0

        chrom_bins = self._child_pack["chrom_bins"]

        for read in self.__read_generator__(unsorted_paths):
            
            loner_type = self.__get_loner_type__(read,chrom_bins)
            loner_total += 1
            write_loner(loner_holder,loner_file_counts,loner_type,read)

        loner_paths = self.__get_paths_and_close__(loner_holder)

        match_maker_results = []
        for loner_type,paths in loner_paths.items():
            for path in paths:
                match_maker_results.append(self.__start_matchmaker_task__([path],loner_type,-1))

        del loner_paths,loner_file_counts,loner_holder
        return PrimaryResults(match_maker_results,"primary")

    def __start_matchmaker_task__(self,paths,loner_type,level):
        return self.match_handler.run(paths,loner_type,level)
        
    def __read_generator__(self,unsorted_paths):
        for path in unsorted_paths:
            loner_object = pysam.AlignmentFile(path,"rb")
            for read in loner_object.fetch(until_eof=True):
                yield read
            loner_object.close()
            os.remove(path)
        gc.collect()

    def __get_reference_id_name__(self,read,bins):
        if read.reference_id == read.next_reference_id:
            return "XX%dv%d" % (read.reference_id,read.next_reference_id)
        else:
            class_bins = map(lambda x : bins[x],sorted((read.reference_id,read.next_reference_id,)))
            #reads on different chromosome
            return "MM%sv%s" % tuple(class_bins)

    def __get_loner_type__(self,read,bins):
        if not read.is_unmapped and not read.mate_is_unmapped:
            return self.__get_reference_id_name__(read,bins)
        elif read.is_unmapped and read.mate_is_unmapped:
            #both unmapped 
            return "UU"
        else:
            #one unmapped read, one mapped read
            return "UM"
            
    def __get_loner_holder__(self):
        holder = {}
        return holder

    def __get_loner_object__(self,loner_type,unique):
        path = self.__get_loner_temp_path__(loner_type,0,unique="-%d" % (unique,) )
        return pysam.AlignmentFile(path,"wb",header=self._header)

    def __write_loner__(self,loner_objects,loner_file_counts,loner_type,read):
        try:
            loner_path = loner_objects[loner_type][-1].filename
            if loner_file_counts[loner_path] == ((self._max_reads / 2) - 1): #Ensure fewer reads than max
                new_bam_object = self.__get_loner_object__(loner_type,len(loner_objects[loner_type]))
                loner_objects[loner_type][-1].close()
                
                loner_objects[loner_type].append(new_bam_object) 
        except KeyError:
            new_bam_object = self.__get_loner_object__(loner_type,0)
            loner_path = new_bam_object.filename
            loner_objects[loner_type] = [new_bam_object]
        
        loner_file_counts[loner_path] += 1
        loner_objects[loner_type][-1].write(read)

    def __get_paths_and_close__(self,loner_holder):   
        paths = {}
        for loner_type,loner_objects in loner_holder.items():
            paths[loner_type] = []
            for loner_object in loner_objects:
                paths[loner_type].append(loner_object.filename)
            loner_object.close()#This close prevents too many open files
        return paths

class MatchMakerHandler(ChaserHandler):

    def __init__(self,object constants,int max_reads, dict child_pack,int pid):
        super(MatchMakerHandler,self).__init__(constants,max_reads,child_pack, pid)
        self.LonerInfo = namedtuple("LonerInfo","paths level loner_type")

    def run(self,loner_paths,loner_type,level):

        loner_info = self.LonerInfo(loner_paths,level,loner_type)
        
        loner_path = self.__get_loner_temp_path__(loner_info.loner_type,loner_info.level+1)
        loner_object = self.__get_loner_object__(loner_path)

        cdef dict pairs = self.__first_pass__(loner_info)
        cdef dict read_pairs = {}
        cdef dict send_pairs = {}
        cdef int rescued = 0
        cdef int written = 0
        cdef list leftover_paths = []

        for read in self.__read_generator__(loner_info,second_pass=True,
                                            leftover_paths=leftover_paths):
            try:
                status = pairs[read.qname]
                try:
                    read_pairs[read.qname].append(read)
                except KeyError:
                    read_pairs[read.qname] = [read]

                if len(read_pairs[read.qname]) == 2:
                    send_pairs[read.qname] = read_pairs[read.qname]
                    del pairs[read.qname]
                    del read_pairs[read.qname]

                #If over hard limit, send to main stream.
                if len(send_pairs) >= 100000:
                    rescued += self.__launch_child_task__(send_pairs)
                    send_pairs = {}
                    gc.collect()

            except KeyError:
                written += 1
                loner_object.write(read)
        loner_object.close()

        if len(send_pairs) > 0:
            rescued += self.__launch_child_task__(send_pairs)

        del read_pairs   
        del pairs

        gc.collect()

        return_paths = []

        if written == 0:
            os.remove(loner_path)
        else:
            return_paths.append((loner_info.level+1,loner_path))

        if len(leftover_paths) > 0:
            for leftover_path in leftover_paths:
                return_paths.append((loner_info.level,leftover_path))

        loner_pack = MatchMakerResults(loner_type=loner_info.loner_type,results=return_paths,
                                       level=loner_info.level+1,chaser_type="match_maker",
                                       rescued=rescued)
        return loner_pack

    def __first_pass__(self,loner_info):
        cdef dict pairs = {}
        for read in self.__read_generator__(loner_info):
            try:
                status = pairs[read.qname]
                pairs[read.qname] = True
            except KeyError:
                pairs[read.qname] = False
        return self.__prepare_for_second_pass__(pairs)

    def __prepare_for_second_pass__(self,pairs):
        return dict(filter( lambda tup : tup[1], iter(pairs.items()) ))

    def __read_generator__(self,loner_info,second_pass=False,leftover_paths=None):
        count = self._max_reads // len(loner_info.paths)
        count_limit = False
        for path in loner_info.paths:
            loner_object = pysam.AlignmentFile(path,"rb")
            bam_iterator = loner_object.fetch(until_eof=True)
           
            for i,read in enumerate(bam_iterator):
                yield read
                if i == count:
                    count_limit = True
                    break 

            if second_pass:
                if count_limit:
                    count_limit = False
                    leftover_count = 0
                    leftover_object = self.__get_leftover_object__(loner_info,len(leftover_paths))
                    for read in bam_iterator:
                        leftover_count += 1
                        leftover_object.write(read)
                    leftover_object.close()
                    
                    if leftover_count > 0:
                        leftover_paths.append(leftover_object.filename)
                    else:
                        os.remove(leftover_object.filename)
                os.remove(path)
            loner_object.close()
        gc.collect()

    def __launch_child_task__(self,pairs):
        child_pack = self._child_pack
        results = child_pack["chaser_task"].handle_pair_dict(pairs,int("%d%d" % (self.pid,self._record,)))
        child_pack["queue"].put(parabam.core.Package(results=results))
        self._record += 1
        return len(pairs)

    def __get_leftover_object__(self,loner_info,unique):
        path = self.__get_loner_temp_path__(loner_info.loner_type,loner_info.level,"-%d" % (unique,))
        leftover = pysam.AlignmentFile(path,"wb",header=self._header)
        return leftover

    def __get_loner_object__(self,path):
        loner_object = pysam.AlignmentFile(path,"wb",header=self._header)
        return loner_object

#.....happily ever after



from distutils.core import setup
from Cython.Build import cythonize

setup(name='parabam',
      description='Parallel BAM File Analysis',
      version='0.1.7dev',
      author="JHR Farmery",
      license='BSD',
      author_email='jhrf2@cam.ac.uk',
      packages=['parabam', 'parabam.command'],
      package_dir = {'parabam':'parabam','parabam.command':'parabam/command'},
      requires=['cython', 'numpy', 'argparse', 'pysam'],
      scripts=['parabam/bin/parabam'],
      ext_modules=cythonize(('parabam/chaser.pyx','parabam/core.pyx',
                             'parabam/merger.pyx','parabam/command/core.pyx',
                             'parabam/command/stat.pyx',
                             'parabam/command/subset.pyx')))



from setuptools import setup
from setuptools.extension import Extension
from setuptools.command.sdist import sdist as _sdist

cmdclass = {}
ext_modules = [
      Extension("parabam.core", [ "parabam/core.c" ]),
      Extension("parabam.chaser", ["parabam/chaser.c"]),
      Extension("parabam.merger", ["parabam/merger.c"]),
      Extension("parabam.command.subset", ["parabam/command/subset.c"]),
      Extension("parabam.command.stat", ["parabam/command/stat.c"]),
      Extension("parabam.command.core", ["parabam/command/core.c"])] 

class sdist(_sdist):
    def run(self):
        # Make sure the compiled Cython files in the distribution are up-to-date
        from Cython.Build import cythonize
        cythonize(['parabam/chaser.pyx','parabam/core.pyx','parabam/merger.pyx',
                   'parabam/command/core.pyx','parabam/command/stat.pyx',
                   'parabam/command/subset.pyx'])
        _sdist.run(self)
cmdclass['sdist'] = sdist

setup(name='parabam',
      description='Parallel BAM File Analysis',
      version='1.4.1',
      author="JHR Farmery",
      license='GPL',
      author_email = 'jhrf2@cam.ac.uk',
      packages = ['parabam','parabam.command'],
      package_dir = {'parabam':'parabam','parabam.command':'parabam/command'},
      install_requires = ['numpy','argparse','pysam'],
      scripts = ['parabam/bin/parabam'],
      cmdclass = cmdclass,
      ext_modules=ext_modules
      )
